# Copyright (c) 2022 Firebuild Inc.
# All rights reserved.
# Free for personal use and commercial trial.
# Non-trial commercial use requires licenses available from https://firebuild.com.
# Modification and redistribution are permitted, but commercial use of
# derivative works is subject to the same requirements of this license
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# This is the definition of an FBB format used for testing FBB itself.

# This is a Python dictionary, to be read and processed by "generate_fbb".

{
  "tags": [
    ("testing", [
      (REQUIRED, "int",  "reqint"),
      (OPTIONAL, "int",  "optint"),
      (ARRAY,    "int",  "arrint"),
      (REQUIRED, "char", "reqchr"),
      (OPTIONAL, "char", "optchr"),
      (ARRAY,    "char", "arrchr"),
      (REQUIRED, STRING, "reqstr"),
      (OPTIONAL, STRING, "optstr"),
      (ARRAY,    STRING, "arrstr"),
      (ARRAY,    STRING, "arrstr2"),
      (REQUIRED, FBB,    "reqfbb"),
      (OPTIONAL, FBB,    "optfbb"),
      (ARRAY,    FBB,    "arrfbb"),
    ]),

    ("testing2", [
      (OPTIONAL, "int",   "t2"),
    ]),

    ("testing3", [
      (OPTIONAL, "int",   "t3"),
    ]),

    ("testing4", [
      (OPTIONAL, "int",   "t4"),
    ]),
  ]
}
